#Region " Copyright "
'***************************************************************************************************'
'
'   File....... Dynamixel.vb
'   Purpose.... Interface with the Robtis Usb2Dynamixel
'   Author..... CrustCrawler Inc. (Mike Gebhard)
'   Help....... http://forum.crustcrawler.com/
'   Started.... 7/13/2007
'   Updated.... 
'   Version.... 1.0.0.0
'
'   Copyright CrustCrawler, Inc.
'
'
'
'
'***************************************************************************************************
#End Region

#Region " Maintenance Log "
'***************************************************************************************************'
' Description:
' Dynamixel object.  This class conatains methods and properies consistant with the 
'   Dynamixel internal commands for reading and writing to registers.
'
'
' Version           Programmer          Release Date            Details
'--------------------------------------------------------------------------------------------------
' 001               Mike Gebhard        7/13/2007               Original Program
' 002               Mike Gebhard        9/29/2007               Updated the DynamixelRegister Enum
'                                                               to include the register number so that
'                                                               when viewing the enum in intellisence
'                                                               the enum list is in numerical order.
'
'***************************************************************************************************
#End Region

Imports System.IO
Imports System.Text
Imports System.IO.Ports
Imports System.Threading
Imports System.Configuration
Imports System.Collections.Generic
Imports System.Collections.Specialized


Public Class Dynamixel

    'Events
    Public Event DynamixelError As DynamixelErrorHandler
    Public Event DynamixelWrite As DynamixelWriteHandler
    Public Event DynamixelRead As DynamixelReadHandler
    'Public Event DynamixelControlTableChange As DynamixelControlTableChangeHandler

    'Serial port reference
    Private connection As SerialPort
    'Members
    Private _portName As String
    Private _baud As Integer
    Private _readTimeout As Integer
    Private _writeTimeout As Integer
    Private _commandDelay As Integer

    Private response As Boolean = True

    Private Const StartByte As Byte = &HFF


#Region " Enums "

    Public Enum baudrate As Byte
        N1000000 = 1
        N500000 = 3
        N400000 = 4
        N250000 = 7
        N200000 = 9
        N117647_1 = 16
        N57142_9 = 34
        N19230_8 = 103
        N9615_4 = 207
    End Enum

    Public Enum Instruction As Byte
        Ping = 1
        ReadData = 2
        WriteData = 3
        RegWrite = 4
        Action = 5
        Reset = 6
        SyncWrite = 131
    End Enum


    Public Enum DynamixelRegister As Byte
        R00_ModelNumber_L = &H0
        R01_ModelNumber_H = &H1
        R02_FirmwareVersion = &H2
        R03_Id = &H3
        R04_Baud = &H4
        R05_ReturnDelayX2uSec = &H5
        R06_ClockwiseAngleLimit_L = &H6
        R07_ClockwiseAngleLimit_H = &H7
        R08_CounterClockwiseAngleLimit_L = &H8
        R09_CounterClockwiseAngleLimit_H = &H9
        'R10_Blank &HA
        R11_TemperatureLimitC = &HB
        R12_LowVoltageLimit = &HC
        R13_HighVoltageLimit = &HD
        R14_MaxTorque_L = &HE
        R15_MaxTorque_H = &HF
        R16_StatusReturnLevel = &H10
        R17_LedAlarm = &H11
        R18_ShutdownAlarm = &H12
        'R19_Blank &H13
        R20_DownCalibration_L = &H14
        R21_DownCalibration_H = &H15
        R22_UpCalibration_L = &H16
        R23_UpCalibration_H = &H17
        R24_EnableTorque = &H18
        R25_LED = &H19
        R26_ClockwiseComplienceMargin = &H1A
        R27_CounterClockwiseComplienceMargin = &H1B
        R28_ClockwiseComplienceSlope = &H1C
        R29_CounterClockwiseComplienceSlope = &H1D
        R30_GoalPosition_L = &H1E
        R31_GoalPosition_H = &H1F
        R32_Speed_L = &H20
        R33_Speed_H = &H21
        R34_TourqueLimit_L = &H22
        R35_TourqueLimit_H = &H23
        R36_CurrentPosition_L = &H24
        R37_CurrentPosition_H = &H25
        R38_CurrentSpeed_L = &H26
        R39_CurrentSpeed_H = &H27
        R40_CurrentLoad_L = &H28
        R41_CurrentLoad_H = &H29
        R42_CurrentVoltage = &H2A
        R43_CurrentTemperature = &H2B
        R44_RegisteredInstruction = &H2C
        'R45_Blank &H2D
        R46_Moving = &H2E
        R47_Lock = &H2F
        R48_Punch_L = &H30
        R49_Punch_H = &H31
    End Enum

#End Region



#Region " Constructors "
    ''' <summary>
    ''' Constructs the Usb2Dynamixel object with settings found in the appSetting section of the app.config file.
    ''' </summary>
    ''' <remarks>This constructor uses appSetting to configure the Usb2Dynamixel's port, baud, read and write timeouts, and delay between Tx/Rx. </remarks>
    Public Sub New()
        Dim appSettings As NameValueCollection = GetAppSettings()
        If appSettings.Count > 0 Then
            _portName = GetConfigValue(appSettings("portName"), "COM1")
            _baud = GetConfigValue(appSettings("baud"), "1000000")
            _readTimeout = CInt(GetConfigValue(appSettings("readTimeout"), "1000"))
            _writeTimeout = CInt(GetConfigValue(appSettings("writeTimeout"), "1000"))
            _commandDelay = CInt(GetConfigValue(appSettings("commandDelay"), "100"))
        End If
    End Sub

    ''' <summary>
    ''' Constructs the Usb2Dynamixel object with settings found in the appSetting section of the app.config file.
    ''' </summary>
    ''' <remarks>This constructor uses appSetting to configure the Usb2Dynamixel's read and write timeouts, and delay between Tx/Rx. 
    ''' User enters the portName (COM1) and baud (1000000).</remarks>
    Public Sub New(ByVal portName As String, ByVal baud As Integer)
        Dim appSettings As NameValueCollection = GetAppSettings()
        _portName = portName
        _baud = baud
        _readTimeout = CInt(GetConfigValue(appSettings("readTimeout"), "1000"))
        _writeTimeout = CInt(GetConfigValue(appSettings("writeTimeout"), "1000"))
        _commandDelay = CInt(GetConfigValue(appSettings("commandDelay"), "100"))
    End Sub


    'Helper functions
    ' Use the constructor's default if the corresponding appSettings tags is missing in app.config file
    Private Function GetConfigValue(ByVal configValue As String, ByVal defaultValue As String) As String
        If String.IsNullOrEmpty(configValue) Then Return defaultValue
        Return configValue
    End Function

    Private Function GetAppSettings() As NameValueCollection
        Dim appSettings As NameValueCollection = ConfigurationManager.AppSettings
        Return appSettings
    End Function

#End Region


#Region " Properties "

    Public ReadOnly Property PortName() As String
        Get
            Return _portName
        End Get
    End Property


    Public ReadOnly Property Baud() As Integer
        Get
            Return _baud
        End Get
    End Property


    Public ReadOnly Property ReadTimeout() As Integer
        Get
            Return _readTimeout
        End Get
    End Property


    Public ReadOnly Property WriteTimeout() As Integer
        Get
            Return _writeTimeout
        End Get
    End Property


    Public Property CommandDelay() As Integer
        Get
            Return _commandDelay
        End Get
        Set(ByVal value As Integer)
            _commandDelay = value
        End Set
    End Property

#End Region


#Region " Ping "
    ''' <summary>
    ''' Get a status packet from a Dynamixel on the network
    ''' </summary>
    ''' <param name="id">Dynamixel ID</param>
    ''' <returns>Status Packet byte array 0xFF, 0xFF, ID, Length, Error, Params...N, CheckSum</returns>
    ''' <remarks>Use this function to verify Dynamixel connectivety.</remarks>
    Public Function Ping(ByVal id As Integer) As Byte()
        response = True
        'Add start bytes
        Dim u2dCommand As List(Of Byte) = AddStartBytes()

        'Add id
        u2dCommand.Add(id)

        'Calculate length N(params) + 2
        Dim nParams As Byte = 0
        u2dCommand.Add(CalculateLength(nParams))

        'Ping instruction
        u2dCommand.Add(Instruction.Ping)

        'Calculate CheckSum
        u2dCommand.Add(CheckSum(u2dCommand.ToArray))

        'Write command and return response
        Return Write(u2dCommand.ToArray)
    End Function

#End Region


#Region " Read Data "
    ''' <summary>
    ''' Read a single Dynamixel setting.
    ''' </summary>
    ''' <param name="id">Dynamixel ID</param>
    ''' <param name="startAddress">Dynamixel Register Address (Enum)</param>
    ''' <returns>Status Packet byte array 0xFF, 0xFF, ID, Length, Error, Params...N, CheckSum</returns>
    ''' <remarks>This method returns one Dynamixel setting.  If the Dynamixel setting requires a single byte then one byte is returned.  
    ''' If the setting takes two bytes (high and low) then both are returned.</remarks>
    Public Function ReadRegister(ByVal id As Integer, ByVal startAddress As DynamixelRegister) As Byte()
        response = True
        'Read Byte or word from the control table
        Dim bytesToRead As Byte = IIf(startAddress.ToString.Contains("_L"), 2, 1)

        'Add start bytes
        Dim u2dCommand As List(Of Byte) = AddStartBytes()

        'Add id
        u2dCommand.Add(id)

        'Calculate length N(params) + 2
        'TODO: 2 is a constant and should be added to a config file, default value, const,or enum
        Dim nParams As Byte = 2
        u2dCommand.Add(CalculateLength(nParams))

        'Instruction
        u2dCommand.Add(Instruction.ReadData)

        'Register start address
        u2dCommand.Add(startAddress)

        'Bytes to read
        u2dCommand.Add(bytesToRead)

        'Calculate CheckSum
        u2dCommand.Add(CheckSum(u2dCommand.ToArray))

        'Write command and return response
        Return Write(u2dCommand.ToArray)
    End Function

    ''' <summary>
    ''' Read one or more Dynamixel registers.
    ''' </summary>
    ''' <param name="id">Dynamixel ID</param>
    ''' <param name="startAddress">Dynamixel Register Address (Enum)</param>
    ''' <param name="endAddress">Dynamixel Register Address (Enum)</param>
    ''' <returns>Status Packet byte array 0xFF, 0xFF, ID, Length, Error, Params...N, CheckSum.</returns>
    ''' <remarks>Reads Dynamixel registers starting at the starting address specified to the ending address specified.  
    ''' This function uses enumerations to identify Dynamixel address registers.  
    ''' Take a look at the source file or Dynamixel documentation for register names and addresses.</remarks>
    Public Function ReadRegister(ByVal id As Integer, ByVal startAddress As DynamixelRegister, ByVal endAddress As DynamixelRegister) As Byte()
        response = True

        'Read Byte or word from the control table
        Dim bytesToRead As Byte = CByte(endAddress - startAddress + 1)

        'Add start bytes
        Dim u2dCommand As List(Of Byte) = AddStartBytes()

        'Add id
        u2dCommand.Add(id)

        'Calculate length N(params) + 2
        'TODO: The number 2 is a constant and should be added to a config file, default value, const,or enum
        Dim nParams As Byte = 2
        u2dCommand.Add(CalculateLength(nParams))

        'Instruction
        u2dCommand.Add(Instruction.ReadData)

        'Register start address
        u2dCommand.Add(startAddress)

        'Bytes to read
        u2dCommand.Add(bytesToRead)

        'Calculate CheckSum
        u2dCommand.Add(CheckSum(u2dCommand.ToArray))

        'Write command and return response
        Return Write(u2dCommand.ToArray)
    End Function


#End Region


#Region " Write Data "
    ''' <summary>
    ''' Write to Dynamixel's register(s)
    ''' </summary>
    ''' <param name="id">Dynamixel ID</param>
    ''' <param name="startAddress">Dynamixel Register Address (Enum)</param>
    ''' <param name="args">List of parameters to write</param>
    ''' <returns>Status Packet byte array 0xFF, 0xFF, ID, Length, Error, Params...N, CheckSum.</returns>
    ''' <remarks>Write to a Dynamixel's register starting at an address specified address and ending once all parameters in the ParamArray have been loaded.</remarks>
    Public Overloads Function WriteRegister(ByVal id As Integer, ByVal startAddress As DynamixelRegister, ByVal ParamArray args() As Byte) As Byte()
        response = True
        If args.Length > 50 Then
            Throw New ArgumentOutOfRangeException("args()", args, "More than 50 parameters entered.  The Dynamixel has 50 registers")
        End If
        'Count parameters
        Dim paramCounter As Integer = 0

        'Add the 2 start bytes
        Dim u2dCommand As List(Of Byte) = AddStartBytes()

        'Add id byte
        u2dCommand.Add(id)

        'Add length place holder at index 3
        u2dCommand.Add(0)

        'Add Instruction byte
        u2dCommand.Add(Instruction.WriteData)

        'Start address enum
        u2dCommand.Add(startAddress)

        'Count the argument parameters and add the
        '   paramters to the instruction packet
        paramCounter += args.Length
        For Each a As Integer In args
            u2dCommand.Add(a)
        Next

        'Update length parameter
        ' Length = number of data bytes + 3
        u2dCommand(3) = paramCounter + 3

        'Calculate CheckSum byte
        u2dCommand.Add(CheckSum(u2dCommand.ToArray))

        'Write the command and return the response
        Return Write(u2dCommand.ToArray)

    End Function 'WriteRegister

#End Region


#Region " Reg_Write "
    ''' <summary>
    ''' Write to Dynamixel's register(s)
    ''' </summary>
    ''' <param name="id">Dynamixel ID</param>
    ''' <param name="startAddress">Dynamixel Register Address (Enum)</param>
    ''' <param name="Action">If Action is set to true then the Action command must be executed to initiate the command</param>
    ''' <param name="args">List of parameters to write</param>
    ''' <returns>Returns nothing when Action is set to true.  
    ''' If Action is set to false then WriteRegister return a status Packet byte array 
    ''' 0xFF, 0xFF, ID, Length, Error, Params...N, CheckSum</returns>
    ''' <remarks>Use Action set to true if you want to control multiple Dynamixels on the network with minimal latency.
    ''' Each Dynamixel will stay in a standy state until the Action command is executed.  This is a good way to coordinate
    ''' curcial or timing related tasks.  With Action set to false, execution is immediate and does not require the Action command.</remarks>
    Public Overloads Function WriteRegister(ByVal id As Integer, ByVal startAddress As DynamixelRegister, ByVal Action As Boolean, ByVal ParamArray args() As Byte) As Byte()

        'Count parameters
        Dim paramCounter As Integer = 0

        'Add the 2 start bytes
        Dim u2dCommand As List(Of Byte) = AddStartBytes()

        'Add id byte
        u2dCommand.Add(id)

        'Add length place holder at index 3
        u2dCommand.Add(0)

        'Add Instruction
        ' RegWrite writes to the dynamixel control table and keeps the Dynamixel
        '  network in stand-by mode until the action instruction is sent.
        '  No response packet is returned
        '
        ' WriteData writtes values to the Dynamixel control table and is
        '   executed immediatley.  Returns a response packet
        '
        If Action Then
            response = False
            u2dCommand.Add(Instruction.RegWrite)
        Else
            response = True
            u2dCommand.Add(Instruction.WriteData)
        End If

        'Start address enum
        u2dCommand.Add(startAddress)

        'Count the argument parameters and add the
        '   paramters to the instruction packet
        paramCounter += args.Length
        For Each a As Integer In args
            u2dCommand.Add(a)
        Next

        'Update length parameter
        ' Length = number of data bytes + 3
        u2dCommand(3) = paramCounter + 3

        'Calculate CheckSum byte
        u2dCommand.Add(CheckSum(u2dCommand.ToArray))

        'Write the command and return the response
        Return Write(u2dCommand.ToArray)

    End Function 'WriteRegister

#End Region


#Region " Action "
    ''' <summary>
    ''' Trigger an action
    ''' </summary>
    ''' <returns>Nothing</returns>
    ''' <remarks>Use this command to execute commands registered in the Dynamixel network.</remarks>
    Public Function Action() As Byte()
        response = False
        'Add start bytes
        Dim u2dCommand As List(Of Byte) = AddStartBytes()

        'Add brodcast Id 0xFE
        u2dCommand.Add(&HFE)

        'Calculate length N(params) + 2
        Dim nParams As Byte = 0
        u2dCommand.Add(CalculateLength(nParams))

        'Add action instruction
        u2dCommand.Add(Instruction.Action)

        'Calculate CheckSum
        u2dCommand.Add(CheckSum(u2dCommand.ToArray))

        'Write command and return response
        Return Write(u2dCommand.ToArray)
    End Function 'Action

#End Region


#Region " Reset "
    ''' <summary>
    ''' Reset the dynamixel to factory
    ''' </summary>
    ''' <param name="id">Dynamixel ID</param>
    ''' <returns>Status Packet byte array 0xFF, 0xFF, ID, Length, Error, Params...N, CheckSum.</returns>
    ''' <remarks>USe this method to reset your Dynamixel back to factory settings.  The ID is reset to 1.</remarks>
    Public Function Reset(ByVal id As String) As Byte()
        response = True
        'Add start bytes
        Dim u2dCommand As List(Of Byte) = AddStartBytes()

        'Add dynamixel id
        u2dCommand.Add(id)

        'Calculate length N(params) + 2
        Dim nParams As Byte = 0
        u2dCommand.Add(CalculateLength(nParams))

        'Add action instruction
        u2dCommand.Add(Instruction.Reset)

        'Calculate CheckSum
        u2dCommand.Add(CheckSum(u2dCommand.ToArray))

        'Write command and return response
        Return Write(u2dCommand.ToArray)
    End Function

#End Region


#Region " Sync Write "

    ''' <summary>
    ''' Send instructs to all Dynamixels on the network (Overloaded)
    ''' </summary>
    ''' <param name="startAddress">Dynamixel Register Address (Enum)</param>
    ''' <param name="numberOfDynamixels">Number of Dynamixels on the network</param>
    ''' <param name="args">List of parameters to write to the Dynamixel network</param>
    ''' <returns>Nothing</returns>
    ''' <remarks>Use this method to synchronize movements of dynamixels on the network.  
    ''' The SyncWrite write command uses the broadcast ID so each Dynamixel instruction 
    ''' packet must be identical. </remarks>
    Public Function SyncWrite(ByVal startAddress As DynamixelRegister, ByVal numberOfDynamixels As Byte, ByVal ParamArray args() As Byte) As Byte()
        'No response expected from the Dynamixel network
        response = False

        Dim N As Byte = numberOfDynamixels

        'Check if the number of bytes to write is divisible by the number of Dynamixels on the network.
        If (args.Length Mod N = 1) Then
            Throw New ArgumentException("Argument length and Dyanmixel register length do not match", New System.Exception())
        End If

        'Determine the number of bytes to write to each Dynamixel
        '   The ID parameter is not part of the data length so subtact one.
        Dim L As Integer = (args.Length / N) - 1

        'Add start bytes
        Dim u2dCommand As List(Of Byte) = AddStartBytes()

        'Add brodcast Id 0xFE
        u2dCommand.Add(&HFE)

        ' Calculate length =((L+1)*N)+4  - see the Dynamixel manual.
        '   Where L is data length (number of bytes) written to each Dynamixel
        '   and N is the number if Dynamixels on the network
        ' The length index is 3
        u2dCommand.Add(((L + 1) * N) + 4)

        'Add SyncWrite instruction
        u2dCommand.Add(Instruction.SyncWrite)

        'Start address
        u2dCommand.Add(startAddress)
        'Length of bytes to write to a single Dynamixel
        u2dCommand.Add(L)

        'Add parameter bytes
        For Each b As Byte In args
            u2dCommand.Add(b)
        Next

        'Calculate CheckSum
        u2dCommand.Add(CheckSum(u2dCommand.ToArray))

        'Write command and return response
        Return Write(u2dCommand.ToArray)

    End Function

#End Region


#Region " Helper Methods "

    Private Function AddStartBytes() As List(Of Byte)
        Dim l As New List(Of Byte)
        l.Add(StartByte)
        l.Add(StartByte)
        Return l
    End Function


    Private Function CalculateLength(ByVal nParams As Byte) As Byte
        If nParams > 253 Then
            Throw New ArgumentOutOfRangeException("nParams", nParams, "Byte overflow error")
        End If
        Return CByte(nParams + 2)
    End Function


    'Helper function splits integer into low/high bytes
    Private Function SplitInteger(ByVal arg As Integer) As Byte()
        Dim LowHigh(1) As Byte
        LowHigh(0) = CByte((arg And &HFF))
        LowHigh(1) = CByte((arg And &HFF00) / &H100)
        Return LowHigh
    End Function


    Private Function CheckSum(ByVal u2dCommand As Byte()) As Byte
        Dim temp As Integer = 0
        Dim cs As Byte = 0
        Dim i As Integer = 0

        'Skip the two start byte in the check sum calculation
        For i = 2 To u2dCommand.Length - 1
            temp += u2dCommand(i)
        Next
        Return CByte(Not temp And &HFF)
    End Function

#End Region


#Region " Port Functions "
    ''' <summary>
    ''' Open the serial port
    ''' </summary>
    ''' <remarks>Read and Write timeout are configured in the app.config file</remarks>
    Public Sub OpenPort()
        Try
            connection = My.Computer.Ports.OpenSerialPort(_portName, _baud)
            connection.WriteTimeout() = _writeTimeout
            connection.ReadTimeout() = _readTimeout
        Catch ex As Exception
            Throw New Exception("Error in Usb2Dynamixel.OpenPort ", ex)
        End Try
    End Sub 'OpenPort
    ''' <summary>
    ''' Close the serial port
    ''' </summary>
    ''' <remarks></remarks>

    Public Sub ClosePort()
        Try
            If Not connection Is Nothing AndAlso connection.IsOpen Then
                connection.Close()
            End If
        Catch ex As Exception
            Throw New Exception("Error in Usb2Dynamixel.ClosePort ", ex)
        End Try
    End Sub 'ClosePort

    ''' <summary>
    ''' Check if the serial port is open
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function PortIsOpen() As Boolean
        If Not connection Is Nothing Then Return connection.IsOpen
        Return False
    End Function

    ''' <summary>
    ''' Write byte stream to Dynamixel network
    ''' </summary>
    ''' <param name="bytesToWrite">Byte stream</param>
    ''' <returns>Status packet</returns>
    ''' <remarks>This function probably should be private but I made it available for testing formated commands that are send directly to the Dynamixel network.</remarks>
    Private Function Write(ByVal bytesToWrite() As Byte) As Byte()
        Try
            If Not connection Is Nothing AndAlso connection.IsOpen Then
                'Raise write event
                OnDynamixelWrite(New DynamixelWriteArgs(bytesToWrite))
                'Execute write command
                connection.Write(bytesToWrite, 0, bytesToWrite.Length)
                Thread.Sleep(_commandDelay)
                'Response
                Return ReadBuffer()
            Else
                Throw New IOException(String.Format("Error in Usb2Dynamixel.Write {0} is closed ", _portName))
                'Return Nothing
            End If
        Catch ex As IOException
            Throw New IOException("IOException in Usb2Dynamixel.Write", ex)
        Catch ex As Exception
            Throw New Exception("Error in Usb2Dynamixel.Write ", ex)
        End Try
        Return Nothing
    End Function


    Private Function ReadBuffer() As Byte()
        Try
            If Not connection Is Nothing AndAlso connection.IsOpen Then
                If response Then
                    If connection.BytesToRead > 0 Then
                        Dim buff(connection.BytesToRead - 1) As Byte
                        connection.Read(buff, 0, connection.BytesToRead)
                        'Raise read port event
                        OnDynamixelRead(New DynamixelReadArgs(buff))
                        'Unit Test errors
                        'buff(4) = (&H80)
                        'buff(4) = (&H40)
                        'buff(4) = (&H20)
                        'buff(4) = (&H10)
                        'buff(4) = (&H8)
                        'buff(4) = (&H4)
                        'buff(4) = (&H2)
                        'buff(4) = (&H1)
                        'buff(4) = (&HC1)
                        'Check for errors in the return packet
                        If Not buff Is Nothing AndAlso buff.Length > 0 Then
                            ErrorCheck(buff(4))
                        Else
                            'No packet returned = &H80
                            ErrorCheck(&H80)
                        End If
                        Return buff
                    Else
                        'No packet returned  = &H80
                        ErrorCheck(&H80)
                    End If
                End If
            End If

        Catch ex As Exception
            Throw New IOException("Error reading serial port buffer - Usb2Dynamixel.ReadPort ", ex)
        End Try
        Return Nothing
    End Function 'Read

#End Region


#Region " Events "

    'Raise Events
    Private Sub ErrorCheck(ByVal errorCode As Integer)
        If errorCode > 0 Then
            OnDynamixelError(New DynamixelErrorArgs(errorCode))
        End If
    End Sub 'ErrorCheck

    Private Sub OnDynamixelError(ByVal e As DynamixelErrorArgs)
        RaiseEvent DynamixelError(Me, e)
    End Sub 'OnDynamixelError

    Private Sub OnDynamixelWrite(ByVal e As DynamixelWriteArgs)
        RaiseEvent DynamixelWrite(Me, e)
    End Sub 'OnDynamixelWrite

    Private Sub OnDynamixelRead(ByVal e As DynamixelReadArgs)
        RaiseEvent DynamixelRead(Me, e)
    End Sub 'OnDynamixelRead

#End Region


#Region " Overrides "

    Public Overrides Function ToString() As String
        Return String.Format("{0}{1}bps", _portName & vbCrLf, _baud)
    End Function 'ToString

#End Region


#Region " Scan "

    ''' <summary>
    ''' Scan the Dynamixel network and return the ID of all attached Dynamixels 
    ''' </summary>
    ''' <returns>List(Of Integer)</returns>
    ''' <remarks>Overloaded - Uses the ping method</remarks>
    Public Function Scan() As List(Of Byte)
        Return DoScan(0, 253)
    End Function


    ''' <summary>
    ''' Scan the Dynamixel network and return the ID of all attached Dynamixels 
    ''' </summary>
    ''' <param name="endScan">Search from 0 to endScan</param>
    ''' <returns>List(Of Integer)</returns>
    ''' <remarks>Overloaded - Uses the ping method</remarks>
    Public Function Scan(ByVal endScan As Integer) As List(Of Byte)
        If endScan > 254 OrElse endScan < 0 Then
            Throw New OverflowException("The endScan value must be less than 254 and greater than -1")
        End If
        Return DoScan(0, endScan)
    End Function


    ''' <summary>
    ''' Scan the Dynamixel network and return the ID of all attached Dynamixels
    ''' </summary>
    ''' <param name="startScan">Start at ID equals startScan</param>
    ''' <param name="endScan">End at ID equals endScan</param>
    ''' <returns>List(Of Integer)</returns>
    ''' <remarks>Overloaded - Uses the ping method</remarks>
    Public Function Scan(ByVal startScan As Integer, ByVal endScan As Integer) As List(Of Byte)
        If endScan > 254 OrElse endScan < 0 Then
            Throw New OverflowException("The endScan value must be less than 254 and greater than -1")
        End If
        If startScan > 254 OrElse startScan < 0 Then
            Throw New OverflowException("The startScan value must be less than 254 and greater than -1")
        End If
        If startScan > endScan Then
            Throw New OverflowException("StartScan must be less than endScan")
        End If
        Return DoScan(startScan, endScan)
    End Function

    'Helper function returns Dynamixel IDs
    Private Function DoScan(ByVal startScan As Integer, ByVal endScan As Integer) As List(Of Byte)
        Dim ids As New List(Of Byte)
        Dim response As Byte() = Nothing

        For id As Integer = startScan To endScan
            response = Ping(id)
            If response IsNot Nothing Then
                ids.Add(response(2))
            End If
        Next
        Return ids
    End Function

#End Region


#Region " LED "
    ''' <summary>
    ''' Turn the Dynamixel LED off and on.
    ''' </summary>
    ''' <param name="id">Dynamixel ID</param>
    ''' <param name="on">LED status</param>
    ''' <returns>Response from Dyanmixel</returns>
    ''' <remarks>Use the Led command to troubleshoot Dynamixel communication issues.</remarks>
    Public Function Led(ByVal id As Integer, ByVal [on] As Boolean) As Byte()
        If [on] Then
            Return WriteRegister(id, DynamixelRegister.R25_LED, CByte(1))
        Else
            Return WriteRegister(id, DynamixelRegister.R25_LED, CByte(0))
        End If
    End Function

#End Region


#Region " Dump "

    Public Function Dump(ByVal id As Byte) As DynamixelControlTable
        Return GetDynamixelControlTable(id)
    End Function


    Private Function GetDynamixelControlTable(ByVal id As Integer) As DynamixelControlTable
        Dim controlTableArray As Byte() = ReadRegister(id, DynamixelRegister.R00_ModelNumber_L, DynamixelRegister.R49_Punch_H)
        If controlTableArray IsNot Nothing Then
            Dim cta As New DynamixelControlTable(controlTableArray)
            AddHandler cta.DynamixelControlTableChange, AddressOf ControlTableChangeHandler
            Return cta
        Else
            Return Nothing
        End If
    End Function


    Public Sub ControlTableChangeHandler(ByVal sender As Object, ByVal e As DynamixelControlTableArgs)
        WriteRegister(e.ID, e.Register, e.Args)
    End Sub


#End Region



    Protected Overrides Sub Finalize()
        MyBase.Finalize()
    End Sub

    Public Overrides Function Equals(ByVal obj As Object) As Boolean
        Return MyBase.Equals(obj)
    End Function

End Class 'Usb2Dynamixel


#Region " Old Stuff "

'Old calculations for using an integer
'For i As Integer = 0 To args.Length - 1 Step argBounds
'    'Add Dynamixel Id
'    u2dCommand.Add(args(i) And &HFF)
'    paramCounter += 1
'    For y As Integer = 1 To argBounds - 1
'        u2dCommand.Add(SplitInteger(args(i + y))(0))
'        paramCounter += 1
'        u2dCommand.Add(SplitInteger(args(i + y))(1))
'        paramCounter += 1
'    Next
'Next

'''' <summary>
'''' Send instructs to all Dynamixels on the network (Overloaded).
'''' </summary>
'''' <param name="startAddress">Dynamixel Register Address (Enum)</param>
'''' <param name="endAddress">Dynamixel Register Address (Enum)</param>
'''' <param name="args">List of parameters to write to the Dynamixel Network</param>
'''' <returns>Nothing</returns>
'''' <remarks>Use this method to synchronize movements of all dynamixels on the network.  
'''' Communication latency is decreased by using the SyncWrite method since the Dynamixels 
'''' are not sending a response packect.  This command uses the broadcast ID.  Each Dynamixel
'''' on the network buffers the instruction and waits for the action command.  The action command  
'''' executes the buffered instructions.</remarks>
'Public Function SyncWrite(ByVal startAddress As DynamixelRegister, ByVal endAddress As DynamixelRegister, ByVal ParamArray args() As Byte) As Byte()
'    response = False

'    'Count parameters
'    'Dim paramCounter As Integer = 0

'    'Determine the number of bytes to write to each Dynamixel
'    Dim L As Integer = (endAddress - startAddress)

'    ''Check if the number of bytes to write is divisible by the number of destination registers.
'    ''   Also make sure that the bytes to write is larger then the destination registers.
'    'If (args.Length Mod L = 1) OrElse (args.Length < L) Then
'    '    Throw New ArgumentException("Argument length and Dyanmixel register length do not match", New System.Exception())
'    'End If

'    ' The number of Dynamixels is determined by the number of arguments divided by
'    '   number of destination registers.
'    Dim N As Byte = (args.Length / L) - 1

'    'Add start bytes
'    Dim u2dCommand As List(Of Byte) = AddStartBytes()

'    'Add brodcast Id 0xFE
'    u2dCommand.Add(&HFE)

'    ' Calculate length =((L+1)*N)+4  - see the Dynamixel manual.
'    '   Where L is data length (number of bytes) written to each Dynamixel
'    '   and N is the number if Dynamixels on the network
'    ' The length index is 3
'    u2dCommand.Add(((L + 1) * N) + 4)

'    'Add SyncWrite instruction
'    u2dCommand.Add(Instruction.SyncWrite)

'    'Start address
'    u2dCommand.Add(startAddress)
'    'Length of bytes to write to a single Dynamixel
'    u2dCommand.Add(L)

'    'Add parameter bytes
'    For Each b As Byte In args
'        u2dCommand.Add(b)
'    Next


'    'Calculate CheckSum
'    u2dCommand.Add(CheckSum(u2dCommand.ToArray))

'    'Write command and return response
'    Return Write(u2dCommand.ToArray)

'End Function 'SyncWrite


#End Region